/** Parent class of all TableOfVars like object that can contain variables */
function TableOfVarsAbstract(
    parentTable     // Optional: TableOfVarsAbstract instance containing this instance (if any)
) {
    if (parentTable)
        this.parentTable = parentTable;
    this.vars = [];
}

/** TableOfVarsAbstract instance containing this instance (if any) */
TableOfVarsAbstract.prototype.parentTable = null;

/** All instances maintaining a single variable */
TableOfVarsAbstract.prototype.vars = [];

/** Available types of variables */
Object.defineProperty(TableOfVarsAbstract.prototype, 'varTypes', {
    enumerable: true,
    get() { return (this.parentTable ? this.parentTable.varTypes : []); }
});

/** Available classes: Array of objects like { className: 'Name of class', abstact: 'true if there cannot be an instance of this class', isa: 'Array of class names of direct/not direct parent classes', atts: 'Array of all attributes (like vars)' } */
Object.defineProperty(TableOfVarsAbstract.prototype, 'classes', {
    enumerable: true,
    get() { return (this.parentTable ? this.parentTable.classes : []); },
    set(v) { if (this.parentTable) this.parentTable.classes = v; }
});

/** All instances maintining a single object instance or an array */
Object.defineProperty(TableOfVarsAbstract.prototype, 'instances', {
    enumerable: true,
    get() { return (this.parentTable ? this.parentTable.instances : []); },
    set(arr) { if (this.parentTable) this.parentTable.instances = arr; }
});

/** All scopes */
Object.defineProperty(TableOfVarsAbstract.prototype, 'scopes', {
    enumerable: true,
    get() { return (this.parentTable ? this.parentTable.scopes : []); },
    set(arr) { if (this.parentTable) this.parentTable.scopes = arr; }
});

/** DOM element containing data */
TableOfVarsAbstract.prototype.nodeData = null;

/** Function to call whenever adding of variable failed */
TableOfVarsAbstract.prototype.onaddfailed = function(msg) {
    if (this.parentTable)
        this.parentTable._addfailed(msg);
}

/** May variables be overloaded? */
Object.defineProperty(TableOfVarsAbstract.prototype, 'enableOverload', {
    enumerable: true,
    get() { return (this.parentTable ? this.parentTable.enableOverload : false); },
    set(v) { if (this.parentTable) this.parentTable.enableOverload = v; }
});

/**
* Converts integer to string representing an address in memory
* @return String    Representation of address in memory
*/
TableOfVarsAbstract.prototype.pointerToAddress = function(
    value,          // Number to convert
    instanceMapping // Optional: Object mapping saved memory pointer to actual pointers
) {
    if (Number.isFinite(value) && (instanceMapping instanceof Object))
        value = Number.isFinite(instanceMapping[value]) ? instanceMapping[value] : value;
    if (!Number.isFinite(value))
        return 'null';
    var hexString = value.toString(16);
    var paddedString = hexString.padStart(4, '0');
    return ((paddedString.length % 2) ? '@0' : '@') + paddedString;
}

/**
* Converts a string representation of address in Memory to index of first byte
* @return Number    Number of first byte used in memory (or null)
*/
TableOfVarsAbstract.prototype.addressToPointer = function(
    address
) {
    const pointer = parseInt(address.substring(1), 16);
    return Number.isFinite(pointer) ? pointer : null;
}

/**
* Create DOM node representing number
* @return Node      DOM node beeing type cell's content
*/
TableOfVarsAbstract.prototype.createCellContentNumber = function(
    doc,        // Document instance
    number     // Number to display
) {
    const nodeCode = doc.createElement('code');
    ScoTools.createHtml('span', nodeCode, { 'class': 'address' }, this.pointerToAddress(number));
    return nodeCode;
}

/**
* Create name cells's DOM node
* @return Node      DOM node beeing name cell's content
*/
TableOfVarsAbstract.prototype.createCellContentName = function(
    doc,        // Document instance
    name        // Variable's name to display or DOM node
) {
    if (name instanceof Node)
        return name;
    if (!name && !Number.isFinite(name))
        return null;
    if (Number.isInteger(name))
        return this.createCellContentNumber(doc, name);
    const nodeCode = doc.createElement('code');
    if (name == 'null')
        ScoTools.createHtml('span', nodeCode, { 'class': 'hljs-number' }, 'null');
    else
        ScoTools.createHtml('span', nodeCode, { 'class': 'hljs-variable' }, name);
    return nodeCode;
}

/**
* Create type cells's DOM node
* @return Node      DOM node beeing type cell's content
*/
TableOfVarsAbstract.prototype.createCellContentType = function(
    doc,        // Document instance
    type,       // Variable's (base) type to display
    dimensions  // Array of given dimension if not zero
) {
    if (type instanceof Node)
        return type;
    if (!type)
        return null;
    const nodeCode = doc.createElement('code');
    ScoTools.createHtml('span', nodeCode, { 'class': 'hljs-type' }, type);
    if (dimensions)
        nodeCode.appendChild(doc.createTextNode('[]'.repeat(dimensions)));
    return nodeCode;
}

/**
* Create value cells's DOM node
* @return Node      DOM node beeing value cell's content
*/
TableOfVarsAbstract.prototype.createCellContentValue = function(
    doc,        // Document instance
    input       // Element for editing value
) {
    if (input instanceof Node)
        return input;
    return null;
}

/**
* Create edit cells's DOM node
* @return Node      DOM node beeing edit cell's content
*/
TableOfVarsAbstract.prototype.createCellContentEdit = function(
    doc,            // Document instance
    removeCallback, // Optional: Function to call when remove button is pressed
    typeDesc        // Optional: Name of item type
) {
    if (!(removeCallback instanceof Function))
        return null;
    const nodeBtn = doc.createElement('button');
    nodeBtn.title = (typeDesc ? typeDesc : 'Variable') + ' löschen';
    nodeBtn.textContent = '❌';
    nodeBtn.addEventListener('click', removeCallback);
    return nodeBtn;
}

/**
* Create button to start scope
* @return Node      DOM node beeing edit cell's content
*/
TableOfVarsAbstract.prototype.createScopeStartBtn = function(
    doc,            // Document instance
    pointer         // Pointer of relevant instance (Number) or null
) {
    const nodeBtn = doc.createElement('button');
    nodeBtn.title = 'Sichtbarkeitsbereich beginnen ' + ((pointer === null) ? '(ohne Wert für this)' : 'mit diesem Objekt als this');
    nodeBtn.textContent = '🚪 Betreten';
    nodeBtn.addEventListener('click', () => this.addScope(pointer));
    return nodeBtn;
}

/**
* Create table row
* @return HTMLTableRowElement       Created DOM node
*/
TableOfVarsAbstract.prototype.createTr = function(
    ...cells        // Strings to display or DOM nodes to display in the corresponding table cell
) {
    const doc = this.getDocument();
    const tr = doc.createElement('tr');
    cells = cells.slice(0, 4);
    while (cells.length < 4)
        cells.push(null);
    cells.forEach(content => {
        const td = ScoTools.createHtml('td', tr);
        if (content)
            td.appendChild(content);
    });
    return tr;
}

/**
* Create table row representing a single variable
* @return HTMLTableRowElement       Created DOM node
*/
TableOfVarsAbstract.prototype.createTrVar = function(
    name,           // Variable's name to display
    type,           // Variable's type to display
    input,          // Element for editing value
    removeCallback  // Optional: Function to call when remove button is pressed
) {
    const doc = this.getDocument();
    return this.createTr(
        this.createCellContentName(doc, name),
        this.createCellContentType(doc, type),
        this.createCellContentValue(doc, input),
        this.createCellContentEdit(doc, removeCallback)
    );
}

/**
* Get Document instance in with to establish DOM nodes
* @return Document      Responsible Document instance
*/
TableOfVarsAbstract.prototype.getDocument = function() {
    return ((this.nodeData instanceof Array) ? this.nodeData[0] : this.nodeData).ownerDocument;
}

/**
* Add a single row for an (sub-)variable
* @return HTMLTableRowElement       Created table row
*/
TableOfVarsAbstract.prototype.addTrVar = function(
    name,           // Variable's name to display
    type,           // Variable's type to display
    input,          // Element for editing value
    removeCallback  // Optional: Function to call when remove button is pressed
) {
    const tr = this.createTrVar(name, type, input,removeCallback);
    if (this.nodeData instanceof Array) {
        const last = this.nodeData[this.nodeData.length - 1];
        last.parentNode.insertBefore(tr, last.nextSibling);
        this.nodeData.push(tr);
    } else
        this.nodeData.appendChild(tr);
    return tr;
}

/**
* Add a single row containing general information on object's instance or array and add managing object to this.instances
* @return HTMLTableRowElement       Created table row
*/
TableOfVarsAbstract.prototype.registerInstance = function(
    instance,       // TableOfVarsInstance object representing a class's instance or TableOfVarsArray object representing an array
    type,           // null, class name (String) or array name (DOM Node) to display
    dimensions     // Optional: Dimension of array (if not zero or undefined)
) {
    if (this.parentTable)
        return this.parentTable.registerInstance(instance, type, dimensions);
    const doc = this.getDocument();
    const tr = this.createTr(
        this.createCellContentName(doc, (type === null) ? 'null' : instance.pointer),
        (type === null) ? null : this.createCellContentType(doc, type, dimensions),
        (dimensions && (type !== null)) ? null : this.createScopeStartBtn(doc, (type === null) ? null : instance.pointer),
        (type === null) ? null : this.createCellContentEdit(doc, () => instance.tryRemove(), (instance instanceof TableOfVarsArray) ? 'Array' : 'Objekt')
    );
    let insertAtIndex = Number.isFinite(instance.pointer) ? this.instances.filter(obj => Number.isFinite(obj.pointer) ? (obj.pointer < instance.pointer) : true).length : 0;
    if (insertAtIndex < this.instances.length)
        this.nodeData.insertBefore(tr, this.instances[insertAtIndex].nodeData[0]);
    else if (this.instances.length) {
        const nextInstance = this.instances[this.instances.length - 1];
        this.nodeData.insertBefore(tr, nextInstance.nodeData[nextInstance.nodeData.length - 1].nextSibling);
    } else
        this.nodeData.insertBefore(tr, this.nodeData.firstChild);
    this.instances.push(instance);
    this.instances.sort((a, b) => {
        if (!Number.isFinite(a.pointer))
            return -1;
        if (!Number.isFinite(b.pointer))
            return +1;
        return a.pointer - b.pointer;
    });
    return tr;
}

/**
* Call this.onaddfailed
* @return undefined
*/
TableOfVarsAbstract.prototype._addfailed = function(
    msg         // Human readable description of problem
) {
    if (!(this.onaddfailed instanceof Function))
        return;
    this.onaddfailed.call(this, msg);
}

/** Array of keyword beeing illegal variable's name */
TableOfVarsAbstract.prototype.forbiddenNames = ["abstract", "assert", "boolean", "break", "byte", "case", "catch", "char", "class", "continue", "const", "default", "do", "double", "else", "enum", "extends", "final", "finally", "float", "for", "goto", "if", "implements", "import", "instanceof", "int", "interface", "long", "module", "native", "new", "package", "private", "protected", "public", "requires", "return", "short", "static", "strictfp", "super", "switch", "synchronized", "this", "throw", "throws", "transient", "try", "var", "void", "volatile", "while", "true", "false", "null"];

/** Allowed characters anywhere within a variable's name */
TableOfVarsAbstract.prototype.nameAllowedCharAnywhere = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz$_';

/** Allowed characters anywhere within a variable's name but at start */
TableOfVarsAbstract.prototype.nameAllowedCharNotStart = '0123456789';

/**
* Add a new variable (if possible)
* @return Boolean       Variable added?
*/
TableOfVarsAbstract.prototype.addVar = function(
    name,           // Variable's name to display
    type,           // Variable's type (matching a this.varTypes[n].typeName)
    value,          // Optional: Value of variable to create
    dimensions,     // Optional: Array of given dimension if not zero
    instanceMapping // Optional: Object mapping saved memory pointer to actual pointers
) {
    if (this.forbiddenNames.indexOf(name) != -1) {
        this._addfailed('Das reservierte Wort „' + name + '“ ist nicht als Variablenname erlaubt');
        return false;
    }
    if (name.length === 0) {
        this._addfailed('Variablenname fehlt');
        return false;
    }
    if (!this.enableOverload && this.vars.reduce((found, variable) => (found || (variable.varName == name)), false)) {
        this._addfailed('Variable „' + name + '“ existiert bereits');
        return false;
    }
    for (let n = 0; n < name.length; ++n) {
        if ((this.nameAllowedCharAnywhere + (n ? this.nameAllowedCharNotStart : '')).indexOf(name.charAt(n)) != -1)
            continue;
        this._addfailed('Das Zeichen „' + name.charAt(n) + '“ darf nicht ' + (n ? 'im Variablenname vorkommen' : 'das erste Zeichen des Variablennamen sein'));
        return false;
    }

    const classInfo = this.classes.reduce((prev, curr) => (prev || ((curr.className == type) ? curr : null)), null);
    const varClass = this.varTypes.reduce((prev, curr) => (prev || ((curr.typeName == type) ? curr : null)), null);
    if (!varClass && !classInfo) {
        this._addfailed('Unbekannter Datentyp „' + type + '“ für neue Variable „' + name + '“');
        return false;
    }

    const variable = dimensions ? (new VarTypeReference(this, name, type, dimensions)) : (classInfo ? (new VarTypeReference(this, name, type, 0, classInfo.abstract)) : (new varClass(this, name)));
    if (value !== undefined)
        variable.value = (dimensions || classInfo) ? this.pointerToAddress(value, instanceMapping) : value;
    this.vars.push(variable);
    return true;
}

/**
* Add a new scope to this table
* @return TableOfVarsScope      Added instance
*/
TableOfVarsAbstract.prototype.addScope = function(
    address     // Memory location of instance beeing this
) {
    if (this.parentTable)
        return this.parentTable.addScope(address);
    return null;
}

/**
* Create and append a TBODY element
* @return HTMLTableBodyElement      Created TBODY element
*/
TableOfVarsAbstract.prototype.addTbody = function() {
    if (this.parentTable)
        return this.parentTable.addTbodyScope(scope);
}

/**
* Adds multiple variables
* @return Array     Array of booleans representing success of adding of corresponding variable
*/
TableOfVarsAbstract.prototype.addVars = function(
    variables,      // Array of objects like { key: "Variable's name", value: "Variable's value", type: "Variable's type" } describing all variables
    instanceMapping // Optional: Object mapping saved memory pointer to actual pointers
) {
    if (!(variables instanceof Array))
        return [ false ];
    return variables.map(({ key, type, value, ref, dimensions }) => this.addVar(key, type, (ref !== undefined) ? ref : value, dimensions, instanceMapping));
}

/**
* Remove a variable
* @return undefined
*/
TableOfVarsAbstract.prototype.removeVar = function(
    variable,       // VarType instance to remove
    arrTr           // Optional: Table rows to remove
) {
    if (!(arrTr instanceof Array)) {
        variable.remove();
        return;
    }
    this.vars = this.vars.filter(item => (item != variable));
    arrTr.forEach(tr => {
        if (tr && tr.parentNode)
            tr.parentNode.removeChild(tr);
    });
    if (this.nodeData instanceof Array)
        this.nodeData = this.nodeData.filter(tr => (arrTr.indexOf(tr) == -1));
}

/**
* Remove a scope
* @return undefined
*/
TableOfVarsAbstract.prototype.removeScope = function(
    scope,          // TableOfVarsScope instance to remove
    tbody           // Optional: Element to remove
) {
    if (!tbody) {
        scope.remove();
        return;
    }
    if (tbody.parentNode)
        tbody.parentNode.removeChild(tbody);
    this.scopes = this.scopes.filter(obj => (obj != scope));
}

/**
* Remove an instance
* @return undefined
*/
TableOfVarsAbstract.prototype.removeInstance = function(
    instance,       // TableOfVarsInstance instance to remove
    arrTr           // Optional: Table rows to remove
) {
    if (!(arrTr instanceof Array)) {
        instance.remove();
        return;
    }
    this.instances = this.instances.filter(item => (item != instance));
    arrTr.forEach(tr => {
        if (tr && tr.parentNode)
            tr.parentNode.removeChild(tr);
    });
    const address = this.pointerToAddress(instance.pointer);
    this.getVarsOfType(instance.baseType, instance.dimensions).forEach(variable => variable.removeInstance(address));
}

/**
* Removes all variables (and scopes and instances only if this is root instance)
* @return undefined
*/
TableOfVarsAbstract.prototype.clear = function() {
    this.vars.forEach(variable => variable.remove());
    if (this.parentTable)
        return;
    this.scopes.forEach(scope => scope.remove());
    this.instances.forEach(instance => instance.remove());
}

/**
* Replaces current variables by a new set of variables
* @return Array     Array of booleans representing success of adding corresponding variable
*/
TableOfVarsAbstract.prototype.load = function(
    variables,      // Array of objects like { key: "Variable's name", value: "Variable's value", type: "Variable's type" } describing all variables or result of TableOfVarsAbstract.toJSON
    instanceMapping // Optional: Object mapping saved memory pointer to actual pointers
) {
    this.clear();
    if (!(instanceMapping instanceof Object))
        instanceMapping = {};
    if (variables instanceof Array)
        return this.addVars(variables, instanceMapping);
    let results = [];
    if (variables.instances instanceof Array)
        variables.instances
            .map(data => {
                const obj = (data.instance == 'array')
                    ? this.createInstanceOfArray(data.baseType, data.dimensions, (data.values || data.refs).length, data.pointer)
                    : this.createInstanceOfObject(data.className, data.pointer);
                if (!(obj instanceof TableOfVarsInstancelike))
                    return (() => results.push(false));
                instanceMapping[data.pointer] = obj.pointer;
                return (data.instance == 'array') ? (() => obj.load(data.values || data.refs, instanceMapping)) : (() => obj.load(data.atts, instanceMapping));
            })
            .forEach(fkt => results.push(fkt().reduce((a, b) => (a && b), true)));
    results = results.concat(this.addVars(variables.vars, instanceMapping));
    if (variables.scopes instanceof Array)
        variables.scopes
            .forEach(scope => {
                let pointer = scope.ref;
                if (Number.isFinite(pointer) && Number.isFinite(instanceMapping[pointer]))
                    pointer = instanceMapping[pointer];
                if (this.instances.map(({ pointer }) => pointer).indexOf(pointer) == -1)
                    pointer = null;
                results.push(this.addScope(pointer).load(scope.vars).reduce((a, b) => (a && b), true))
            });
    return results;
}

/**
* Get first position in memory to fit `reserve` bytes
* @return Number    Memory position
*/
TableOfVarsAbstract.prototype.allocateMemory = function(
    reserve     // Number of bytes to reserve
) {
    reserve = Math.max(reserve, 1);
    const allocated = this.instances.map(({ pointer, size }) => ({ pointer, size: (size > 0) ? size : 1 })).filter(({ pointer }) => Number.isFinite(pointer));
    allocated.sort((a, b) => (a.pointer - b.pointer));
    return allocated.reduce((pos, { pointer, size }) => ((pos + reserve <= pointer) ? pos : (pointer + size)), 0);
}

/**
* Check if there is space for `reserve` bytes beginning at memory location `pointer`
* @return Boolean       Is described memory location available?
*/
TableOfVarsAbstract.prototype.canAllocate = function(
    pointer,    // Memory address of first byte
    reserve     // Number of bytes to reserve
) {
    const min = pointer;
    const max = pointer + reserve;
    return this.instances
        .map(({ pointer, size }) => ((max <= pointer) || (min >= pointer + size)) )
        .reduce((a, b) => (a && b), true);
}

/**
* Check if `subClassName` is, implements or extends `parentClassName` according to `this.classes`
* @return Boolean       `subClassName` is, implements or extends `parentClassName`
*/
TableOfVarsAbstract.prototype.isSubclassOf = function(
    subClassName,       // Name of class to check
    parentClassName     // Name of required class
) {
    if (subClassName == parentClassName)
        return true;
    const classInfo = this.classes.reduce((prev, curr) => (prev || ((curr.className == subClassName) ? curr : null)), null);
    return (classInfo && (classInfo.isa instanceof Array) && (classInfo.isa.indexOf(parentClassName) > -1));
}

/**
* Get all instances of baseType (if dimensions == 0) or dimensions dimensional arrays of baseType
* @return Array     Addresses of all instances (sorted, items are strings like '@xxxx')
*/
TableOfVarsAbstract.prototype.getInstancesOf = function(
    baseType,       // Class name or array items' variable type
    dimensions      // Array of given dimension if not zero
) {
    return this.instances
        .filter(obj => (obj instanceof TableOfVarsInstancelike))
        .filter((obj) => (this.isSubclassOf(obj.baseType, baseType) && (obj.dimensions == dimensions)))
        .map(({ pointer }) => this.pointerToAddress(pointer))
        .sort();
}

/**
* Get all VarTypeReference instances of given type within hole table
* @return Array     All found instances
*/
TableOfVarsAbstract.prototype.getVarsOfType = function(
    baseType,       // Class name or array items' variable type
    dimensions      // Array of given dimension if not zero
) {
    return this.parentTable ? this.parentTable.getVarsOfType(baseType, dimensions) : this.getContainedVarsOfType(baseType, dimensions);
}

/**
* Get all VarTypeReference instances of given type within this part of the table
* @return Array     All found instances
*/
TableOfVarsAbstract.prototype.getContainedVarsOfType = function(
    baseType,       // Class name or array items' variable type
    dimensions      // Array of given dimension if not zero
) {
    return this.vars.filter(variable => ((variable instanceof VarTypeReference) && (variable.dimensions == dimensions) && this.isSubclassOf(baseType, variable.baseType)));
}

/**
* Get information on all references to `pointer` within hole table
* @return Array     Information on references beeing objects like { type: "scope|var", obj: "Actual object" }
*/
TableOfVarsAbstract.prototype.getExternalReferences = function(
    pointer         // Memory location of instance in question
) {
    if (this.parentTable)
        return this.parentTable.getExternalReferences(pointer);
    return this.scopes
        .filter(scope => (scope.pointer == pointer))
        .map(scope => ({ type: 'scope', obj: scope }))
        .concat(this.getExternalReferencesContained(this.pointerToAddress(pointer)));
}

/**
* Get information on all references to given address within table part
* @return Array     Information on references beeing objects like { type: "var", obj: "Actual variable" }
*/
TableOfVarsAbstract.prototype.getExternalReferencesContained = function(
    address         // Canonic human readable memory location of instance in question
) {
    return this.vars
        .filter(variable => ((variable instanceof VarTypeReference) && (variable.value == address)))
        .map(variable => ({ type: 'var', obj: variable }));
}

/**
* Creates an instances of baseType (if dimensions == 0) or dimensions dimensional arrays of baseType
* @return String    Human readable address of created instance
*/
TableOfVarsAbstract.prototype.createInstancesOf = function(
    baseType,       // Class name or array items' variable type
    dimensions      // Create array if non zero
) {
    const instance = (dimensions > 0) ? this.createInstanceOfArray(baseType, dimensions, window.prompt('Länge des Arrays:', 5)) : this.createInstanceOfObject(baseType);
    if (!instance)
        return null;
    return this.pointerToAddress(instance.pointer);
}

/**
* Create an instance of `baseType`
* @return TableOfVarsInstance       Created instance
*/
TableOfVarsAbstract.prototype.createInstanceOfObject = function(
    baseType,       // Class name
    pointer         // Optional: Preferred memory location
) {
    if (baseType === null)
        return new TableOfVarsNullInstance(this);
    const classInfo = this.classes.reduce((prev, info) => (info.className == baseType) ? info : prev, null);
    if (!classInfo) {
        this._addfailed('Ein Objekt der unbekannten Klasse „' + baseType + '“ kann nicht erzeugt werden');
        return null;
    }
    const instance = new TableOfVarsInstance(this, classInfo, pointer);
    const address = this.pointerToAddress(instance.pointer);
    this.getVarsOfType(baseType, 0).forEach(variable => variable.addInstance(address));
    return instance;
}

/**
* Create an instance of a `dimensions` dimensional `baseType` array
* @return TableOfVarsArray          Created instance
*/
TableOfVarsAbstract.prototype.createInstanceOfArray = function(
    baseType,       // Array items' variable base type
    dimensions,     // Create array if non zero
    length,         // Number of items
    pointer         // Optional: Preferred memory location
) {
    if (!Number.isInteger(+length) || (length < 0))
        return 'Ein Array der Länge ' + length + ' ist nicht möglich';
    const instance = new TableOfVarsArray(this, baseType, dimensions, +length, pointer);
    const address = this.pointerToAddress(instance.pointer);
    this.getVarsOfType(baseType, dimensions).forEach(variable => variable.addInstance(address));
    return instance;
}

/**
* Convert contained data to JSON
* @return Object    JSON representation of contained data
*/
TableOfVarsAbstract.prototype.toJSON = function() {
    if (!this.parentTable && (this.instances.length + this.scopes.length))
        return {
            instances: this.instances.map(item => item.toJSON()),
            vars: this.vars.map(item => item.toJSON()),
            scopes: this.scopes.map(item => item.toJSON())
        };
    return this.vars.map(item => item.toJSON());
}

/**
* Compares present variables, scopes and instances with a target
* @return Array     Human readable descriptions (strings) of all found differences
*/
TableOfVarsAbstract.prototype.compare = function(
    target,                     // Wanted variables, scopes and instances (like TableOfVarsAbstract.toJSON`s result)
    allowAdditionalInstances    // Not complain about additional instances?
) {
    return (new VarsDiff(target, this.toJSON(), pointer => this.pointerToAddress(pointer))).getDiffs(true, allowAdditionalInstances);
}
